# Copyright Justus Schock.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any, Callable, Optional, Sequence, Tuple, Union

import nibabel as nib
import torch
import torchio as tio
from torch.nn import functional as F
from torchio.transforms.preprocessing.spatial.resample import Resample

from mdlu.data.modality import ImageModality

__all__ = [
    "DefaultPreprocessing",
    "ResampleOnehot",
    "ResampleAndCropOrPad",
    "CropToNonZero",
    "NNUnetNormalization",
    "RescaleIntensityPercentiles",
    "DefaultSpatialAugmentation",
    "CropOrPadPerImage",
]


class DefaultSpatialPreIntensityPreprocessingTransforms(tio.transforms.Compose):
    """The default spatial preprocessing to be running BEFORE the intensity preprocessing.

    Can optionally (by default) include a Cropping to nonzero values, a resampling to the target spacing (for masks
    either in one-hot format or using the nearest neighbor resampling).
    """

    def __init__(
        self,
        resample_onehot: bool,
        target_spacing,
        interpolation: str = "linear",
        crop_to_nonzero: str | bool | None = False,
        num_classes: int | None = None,
    ) -> None:
        """
        Args:
            resample_onehot: Whether to resample masks to the target spacing in onme-hot using the same
                interpolation scheme as for images combined with an argmax or using the nearest neighbor approach.
            target_spacing: The target spacing to resample to.
            interpolation: The interpolation scheme to use when resampling.
            crop_to_nonzero: Whether to crop to nonzero values.
            num_classes: The number of classes to use when resampling masks.

        """

        trafos = []

        if crop_to_nonzero or crop_to_nonzero is None:
            if crop_to_nonzero is not None and not isinstance(crop_to_nonzero, str):
                raise TypeError(
                    "crop_to_nonzero must be a string or None or a value evaluating to False to avoid cropping at all"
                )
            trafos.append(CropToNonZero(crop_to_nonzero))

        if resample_onehot:
            assert num_classes is not None
            trafos.append(
                ResampleOnehot(
                    num_classes=num_classes,
                    target=target_spacing,
                    image_interpolation=interpolation,
                )
            )
        else:
            trafos.append(Resample(target=target_spacing, image_interpolation=interpolation))

        super().__init__(trafos)


class DefaultSpatialPostIntensityPreprocessingTransforms(tio.transforms.Compose):
    """The default spatial preprocessing to be running AFTER the intensity preprocessing.

    Currently only consists of cropping or padding to a given target size
    """

    def __init__(self, target_size: torch.Tensor) -> None:
        """
        Args:
            target_size: The target size to crop or pad to.
        """
        trafos = [CropOrPadPerImage(target_shape=target_size)]
        super().__init__(trafos)


class DefaultIntensityPreprocessing(tio.transforms.Compose):
    """The default intensity preprocessing.

    Includes a rescaling of intensity values to certain percentiles as well as a normalization transform.
    """

    def __init__(
        self,
        target_size,
        modality: int | ImageModality | None = None,
        dataset_intensity_mean: torch.Tensor | None = None,
        dataset_intensity_std: torch.Tensor | None = None,
        norm_trafo: Callable[[tio.data.Subject], tio.data.Subject] | None = None,
    ) -> None:
        """
        Args:
            target_size: The target size to crop or pad to.
            modality: The modality to use.
            dataset_intensity_mean: The mean to use for the intensity normalization.
            dataset_intensity_std: The standard deviation to use for the intensity normalization.
            norm_trafo: A function to apply to the subject after normalization.
        """

        if modality == ImageModality.CT:
            percentiles = (0.5, 99.5)
        else:
            percentiles = (0, 100)

        if norm_trafo is None:
            norm_trafo = NNUnetNormalization(
                target_size=target_size,
                image_modality=modality,
                dataset_mean=dataset_intensity_mean,
                dataset_std=dataset_intensity_std,
            )

        trafos = [RescaleIntensityPercentiles(percentiles=percentiles), norm_trafo]
        super().__init__(trafos)


class DefaultPreprocessing(tio.transforms.Compose):
    """The default preprocessing. Consists of the following tranforms (in Order):

    - Transform all images and label images to a canonical RAS+ orientation
    - Copy Affine from image to others (like labels). There can be inaccuracies using different loaders or annotation tools
        which could cause errors downstream if not unifying now.
    - Cropping to nonzero values (optional)
    - Resampling to the target spacing (for masks: either one-hot or nearest neighbor)
    - Rescaling of intensity values to certain percentiles
    - Normalization of intensity values
    - Cropping or padding to target size
    """

    def __init__(
        self,
        target_spacing: Sequence[float],
        target_size: Sequence[int],
        modality: int | ImageModality | None = None,
        dataset_intensity_mean: torch.Tensor | None = None,
        dataset_intensity_std: torch.Tensor | None = None,
        interpolation: str = "linear",
        affine_key: None | str = "data",
        num_classes: int | None = None,
    ):
        """
        Args:
            target_spacing: The target spacing to resample to.
            target_size: The target size to crop or pad to.
            modality: The modality to use.
            dataset_intensity_mean: The mean to use for the intensity normalization.
            dataset_intensity_std: The standard deviation to use for the intensity normalization.
            interpolation: The interpolation scheme to use when resampling.
            affine_key: The key to use for copying the affine from.
            num_classes: The number of classes to use when resampling masks.

        """
        trafos = [
            tio.transforms.ToCanonical(),
        ]

        if affine_key is not None:
            trafos.append(tio.transforms.preprocessing.CopyAffine(target=affine_key))

        trafos += [
            DefaultSpatialPreIntensityPreprocessingTransforms(
                resample_onehot=True,
                num_classes=num_classes,
                target_spacing=target_spacing,
                interpolation=interpolation,
            ),
            DefaultIntensityPreprocessing(
                modality=modality,
                target_size=target_size,
                dataset_intensity_mean=dataset_intensity_mean,
                dataset_intensity_std=dataset_intensity_std,
            ),
            DefaultSpatialPostIntensityPreprocessingTransforms(target_size=target_size),
        ]

        return super().__init__(trafos)


class ResampleOnehot(Resample):
    """Resample a mask to the target spacing using one-hot encoding and same interpolation scheme as for images.

    .. note::
        For additional arguments have a look at :class:`torchio.transforms.preprocessing.spatial.resample.Resample`.
    """

    def __init__(self, num_classes: int, *args: Any, **kwargs: Any) -> None:
        """
        Args:
            num_classes: The number of classes to use when resampling masks.
        """
        super().__init__(*args, **kwargs)
        self.to_onehot_trafo = tio.transforms.OneHot(num_classes=num_classes)
        self.to_categorical = LabelToCategorical()

    def apply_transform(self, subject: tio.data.Subject) -> tio.data.Subject:
        """Applies the resampling to the whole subject.

        Args:
            subject: The subject to apply the transform to.

        Returns:
            The transformed subject.
        """
        subject = self.to_onehot_trafo(subject)
        temp_scalar_images = []

        for k, v in subject.get_images_dict(intensity_only=False).items():
            if not isinstance(v, tio.data.LabelMap):
                continue
            subject[k] = tio.data.ScalarImage(tensor=v.data, affine=v.affine)
            temp_scalar_images.append(k)
        subject = super().apply_transform(subject)
        for k in temp_scalar_images:
            subject[k] = tio.data.LabelMap(tensor=subject[k].data, affine=subject[k].affine)

        subject = self.to_categorical(subject)
        return subject


class ResampleAndCropOrPad(tio.transforms.Transform):
    """Resampling to a custom target spacing and cropping or padding to a custom target size.

    Handles each image individually since depending on initial shape and spacing the in-between results may differ.
    """

    def __init__(
        self,
        target_spacing: tuple[float, float, float],
        target_size: tuple[int, int, int],
        num_classes: int,
        interpolation: str = "linear",
        **kwargs,
    ) -> None:
        """
        Args:
            target_spacing: The target spacing to resample to.
            target_size: The target size to crop or pad to.
            num_classes: The number of classes to use when resampling masks.
            interpolation: The interpolation scheme to use when resampling.
        """
        super().__init__(**kwargs)

        self._resample = ResampleOnehot(
            num_classes,
            target_spacing,
            interpolation=interpolation,
            p=self.probability,
            copy=self.copy,
            include=self.include,
            exclude=self.exclude,
            keys=None,
            keep=self.keep,
        )
        self._crop_or_pad = tio.transforms.CropOrPad(
            target_size,
            p=self.probability,
            copy=self.copy,
            include=self.include,
            exclude=self.exclude,
            keys=None,
            keep=self.keep,
        )

    def apply_transform(self, subject: tio.data.Subject) -> tio.data.Subject:
        """Applies the resampling and cropping or padding to the whole subject.

        Args:
            subject: The subject to apply the transform to.

        Returns:
            The transformed subject.
        """

        # handle each image in subject individually to compensate
        # for misalignment of images in checks inbetween the steps!
        for k, v in subject.get_images_dict(intensity_only=False).items():
            curr_sub = tio.data.Subject(k=v)
            curr_sub = self._resample(curr_sub)
            curr_sub = self._crop_or_pad(curr_sub)

            subject[k] = curr_sub[k]
        return subject


class CropToNonZero(tio.transforms.Transform):
    """Crops the image to a bounding box containing all nonzero voxels."""

    def __init__(self, key: str | None = None, **kwargs: Any) -> None:
        """
        Args:
            key: The key specifying the image to use.
        """
        super().__init__(**kwargs)
        self.key = key

    @staticmethod
    def crop_to_nonzero(image: tio.data.Image, *additional_images: tio.data.Image, **padding_kwargs) -> None:
        """crops tensor to non-zero and additional tensor to the same part of the tensor if given.

        Args:
            tensor: the tensor to crop
            additional_tensor: an additional tensor to crop to the same region
                as :attr`tensor`
            **padding_kwargs: keyword arguments to controll the necessary padding
        """

        centers, ranges = extract_nonzero_bounding_box_from_tensor(image.tensor[None])

        low = (centers - ranges / 2)[0].cpu().detach().numpy()

        for img in (image, *additional_images):
            result = crop_tensor_to_bounding_box(
                img.tensor[None],
                centers=centers,
                ranges=ranges,
                **padding_kwargs,
            )

            img.set_data(result[0])

            img_affine = img.affine.copy()
            new_origin = nib.affines.apply_affine(img.affine, low)
            img_affine[:3, 3] = new_origin
            img.affine = img_affine

    def apply_transform(self, subject: tio.data.Subject) -> tio.data.Subject:
        """Applies the transform to the whole subject.

        Args:
            subject: The subject to apply the transform to.

        Returns:
            The transformed subject.
        """
        intensity_images = subject.get_images_dict(intensity_only=True)
        all_images = subject.get_images_dict(intensity_only=False)

        if self.key is None:
            images = intensity_images
        else:
            images = {self.key: subject[self.key]}

        assert (
            len(images) == 1
        ), "Multiple images found, this may lead to inconsistent cropping behavior. Pleas specify a reference image key"
        # TODO: Add reference image key to default trafos and dataset

        seg_images = {k: v for k, v in all_images.items() if k not in intensity_images}
        for v in images.values():
            self.crop_to_nonzero(v, *seg_images.values())

        return subject


class NNUnetNormalization(tio.transforms.ZNormalization):
    """Normalization similar to the one prposed in the nnUNet paper.

    - If the image is not a CT and not more than 25% of the image are cropped away,
        use the usual Z-Normalization across the whole image.
    - If the image is not a CT and more than 25% of the iamge are cropped away,
        extract the mean and standard deviation of from the non-cropped parts of the image only and
        perform Z-Normalization using these values.
    - If the image is a CT, use the passed mean and standard deviation values computed over the whole dataset
        and use those for Z-Normalization.
    """

    def __init__(
        self,
        target_size: Sequence[int],
        image_modality: int | ImageModality | None = None,
        dataset_mean: torch.Tensor | None = None,
        dataset_std: torch.Tensor | None = None,
    ):
        """
        Args:
            target_size: The target size to crop or pad to.
            image_modality: The image modality to use.
            dataset_mean: The mean to use for the dataset.
            dataset_std: The standard deviation to use for the dataset.
        """
        super().__init__()

        if isinstance(image_modality, int):
            image_modality = ImageModality(image_modality)
        self.image_modality = image_modality
        self.dataset_mean = dataset_mean
        self.dataset_std = dataset_std
        self.num_target_elems = torch.tensor(target_size).prod()
        self.crop_or_pad_trafo = CropOrPadPerImage(target_shape=target_size)

    def apply_normalization(self, subject: tio.data.Subject, image_name: str, mask: torch.Tensor) -> None:
        """Applies the normalization to the image.

        Args:
            subject: The subject to apply the normalization to.
            image_name: The name of the image to apply the normalization to.
            mask: The mask to use for the normalization.
        """

        # no image modality is given, use default (non-CT) behavior
        if self.image_modality is None or self.image_modality != ImageModality.CT:
            # not more than 25% of the voxels are cropped away
            if torch.tensor(subject.spatial_shape).prod() * 0.75 < self.num_target_elems:
                return super().apply_normalization(subject, image_name, mask)

            cropped_tensor = self.crop_or_pad_trafo(
                tio.data.Subject({image_name: tio.data.ScalarImage(tensor=subject[image_name].data)})
            )[image_name].data
            # masking for foreground
            mask = cropped_tensor > 0
            values = cropped_tensor[mask].float()
            mean = values.mean()
            std = values.std()

        else:
            if self.dataset_mean is None or self.dataset_std is None:
                raise ValueError("dataset mean and std can only be None if the dataset modality is not CT.")

            mean = self.dataset_mean
            std = self.dataset_std

        if std == 0:
            mean = self.dataset_mean
            std = self.dataset_std

        image = subject[image_name]
        tensor = image.data.to(mean)
        tensor -= mean
        tensor /= std
        image.set_data(tensor)


class LabelToCategorical(tio.transforms.preprocessing.label.label_transform.LabelTransform):
    """Converts a One-Hot encoded label to a categorical label."""

    def apply_transform(self, subject: tio.data.Subject) -> tio.data.Subject:
        """Applies the transform to the whole subject.

        Args:
            subject: The subject to apply the transform to.

        Returns:
            The transformed subject.
        """

        for k, v in self.get_images_dict(subject).items():
            if isinstance(v, tio.data.LabelMap):
                v.set_data(v.data.argmax(0).unsqueeze(0))
        return subject


class RescaleIntensityPercentiles(tio.transforms.RescaleIntensity):
    """Rescales the intensity of the image to the given percentiles."""

    def rescale(self, tensor: torch.Tensor, mask: torch.Tensor, image_name: str) -> torch.Tensor:
        """Rescales the intensity of the image to the given percentiles.

        Args:
            tensor: The tensor to rescale.
            mask: The mask to use for the rescaling.
            image_name: The name of the image to rescale.
        """
        if self.percentiles == (0, 100):
            return tensor
        else:
            self.out_max = tensor.max().item()
            self.out_min = tensor.min().item()
            return super().rescale(tensor, mask, image_name)


class DefaultSpatialAugmentation(tio.transforms.Compose):
    """Default spatial augmentation.

    - Approximates a random anisotropy
    - Randomly flips the image
    - Applies a random affine transformation
    - Optionally includes a random elastic deformation
    """

    def __init__(self, image_modality: ImageModality | int, include_deformation: bool = False, p: float = 0.25) -> None:
        """
        Args:
            image_modality: The image modality to use.
            include_deformation: Whether to include elastic deformation.
            p: The probability of applying every single transformation.
        """
        trafos = []
        if image_modality is not None:
            trafos += [
                tio.transforms.RandomAnisotropy(
                    axes=tuple(range(ImageModality.get_dimensionality(image_modality))),
                    p=p,
                ),
                tio.transforms.RandomFlip(tuple(range(ImageModality.get_dimensionality(image_modality))), p=p),
            ]

        trafos.append(tio.transforms.RandomAffine(p=p))

        if include_deformation:
            trafos.append(tio.transforms.RandomElasticDeformation(p=p))

        super().__init__(trafos)


class DefaultIntensityAugmentation(tio.transforms.Compose):
    """Default intensity augmentation.

    - Random Ghosting Artifacts (for MRI only)
    - Random Spiking Artifacts (for MRI only)
    - Random Motion Artifacts
    - Random Bias Fields
    - Random Gamma Transform
    - Random Blurring
    - Random Noise
    - Random Swapping of Patches
    """

    def __init__(self, image_modality: ImageModality | int, p: float = 0.25) -> None:
        """
        Args:
            image_modality: The image modality to use.
            p: The probability of applying every single transformation.

        """
        trafos = []

        if image_modality is not None and image_modality == ImageModality.MR:
            trafos.append(tio.transforms.RandomGhosting(p=p))
            trafos.append(tio.transforms.RandomSpike(p=p))

        trafos += [
            tio.transforms.RandomMotion(p=p),
            tio.transforms.RandomBiasField(p=p),
            tio.transforms.RandomGamma(p=p),
            tio.transforms.RandomBlur(p=p),
            tio.transforms.RandomNoise(p=p),
            tio.transforms.RandomSwap(p=p),
        ]
        super().__init__(trafos)


class DefaultAugmentation(tio.transforms.Compose):
    """Default augmentation.

    - Approximates a random anisotropy
    - Randomly flips the image
    - Applies a random affine transformation
    - Optionally includes a random elastic deformatio
    - Random Ghosting Artifacts (for MRI only)
    - Random Spiking Artifacts (for MRI only)
    - Random Motion Artifacts
    - Random Bias Fields
    - Random Gamma Transform
    - Random Blurring
    - Random Noise
    - Random Swapping of Patches
    """

    def __init__(
        self,
        image_modality: ImageModality | int,
        include_deformation: bool = False,
        spatial_prob: float = 0.25,
        intensity_prob: float = 0.5,
    ) -> None:
        """
        Args:
            image_modality: The image modality to use.
            include_deformation: Whether to include elastic deformation.
            spatial_prob: The probability of applying every single spatial transformation.
            intensity_prob: The probability of applying every single intensity transformation.
        """
        trafos = [
            DefaultSpatialAugmentation(
                image_modality=image_modality,
                include_deformation=include_deformation,
                p=spatial_prob,
            ),
            DefaultIntensityAugmentation(image_modality=image_modality, p=intensity_prob),
        ]
        super().__init__(trafos)


class Transform3DTo2D(tio.transforms.Transform):
    """Transforms a 3D image to a 2D image with the number of channels equal to the depth beforehand."""

    def apply_transform(self, subject: tio.data.Subject) -> tio.data.Subject:
        """Applies the transform to the whole subject.

        Args:
            subject: The subject to apply the transform to.

        Returns:
            The transformed subject.
        """
        for k, v in subject.get_images_dict.items():
            tensor = v.data
            orig_shape = v.data.shape
            tensor = tensor.view(orig_shape[3] * orig_shape[0], orig_shape[1], orig_shape[2], 1)
            v.set_data(tensor)
            subject[f"_{k}_orig_shape"] = orig_shape
        return subject


class Transform2DBackTo3D(tio.transforms.Transform):
    """Transforms a 2D image back to a 3D image with the original shape restored."""

    def apply_transform(self, subject: tio.data.Subject) -> tio.data.Subject:
        """Applies the transform to the whole subject.

        Args:
            subject: The subject to apply the transform to.

        Returns:
            The transformed subject.
        """
        for k, v in subject.get_images_dict.items():
            tensor = v.data
            orig_shape = subject[f"_{k}_orig_shape"]
            tensor = tensor.view(orig_shape)
            v.set_data(tensor)
            subject.pop(f"_{k}_orig_shape")
        return subject


def extract_nonzero_bounding_box_from_tensor(
    tensor: torch.Tensor,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Extracts the bounding box of nonzero elements from a tensor.

    Args:
        tensor: The tensor to extract the bounding box from.

    Returns:
        The bounding box of nonzero elements (as centers and ranges).
    """

    # create empty ranges
    ranges = torch.zeros(
        tensor.size(0),
        tensor.ndim - 2,
        dtype=torch.long,
        device=tensor.device,
    )
    # create empty centers
    centers = torch.zeros(
        tensor.size(0),
        tensor.ndim - 2,
        dtype=torch.long,
        device=tensor.device,
    )
    for idx, _tensor in enumerate(tensor):
        bounds = torch.tensor(
            [(tmp.min(), tmp.max()) for tmp in _tensor.nonzero(as_tuple=True)],
            device=tensor.device,
        )
        # ensure this is an even number
        ranges[idx, :] = (torch.ceil((bounds[:, 1] - bounds[:, 0]).float() / 2) * 2).long()[
            1:
        ]  # extract max range of all channels
        centers[idx, :] = bounds[1:, 0] + ranges[idx] // 2

    return centers, ranges


def crop_tensor_to_bounding_box(
    tensor: torch.Tensor,
    centers: torch.Tensor,
    ranges: torch.Tensor,
    additional_tensor: torch.Tensor | None = None,
    **padding_kwargs,
) -> tuple[torch.Tensor, torch.Tensor] | torch.Tensor:
    """Crops a tensor to a bounding box.

    Args:
        tensor: The tensor to crop.
        centers: The centers of the bounding box.
        ranges: The ranges of the bounding box.
        additional_tensor: An additional tensor to crop with the same bounbding box.
        **padding_kwargs: Additional arguments to pass to the padding function.

    Returns:
        The cropped tensor optionally (if passed) along with the cropped additional tensor.
    """
    max_range = ranges.max(0)[0]
    out = torch.zeros(
        tensor.size(0),
        tensor.size(1),
        *max_range.tolist(),
        device=tensor.device,
        dtype=tensor.dtype,
    )

    if additional_tensor is not None:
        additional_out = torch.zeros(
            additional_tensor.size(0),
            additional_tensor.size(1),
            *max_range.tolist(),
            device=additional_tensor.device,
            dtype=additional_tensor.dtype,
        )

    # calciulate the absolute minimum range
    abs_min_range = (centers - max_range / 2.0).min(0)[0]

    # calculate padding on the first side per dim
    first_paddings = torch.where(abs_min_range < 0, abs_min_range.abs(), torch.zeros_like(abs_min_range))

    # calculate padding on the last side per dim
    last_paddings = torch.clamp(
        (centers + max_range / 2.0).max(0)[0]
        - torch.tensor(tensor.shape[2:], device=centers.device, dtype=centers.dtype),
        0,
    )

    # combine paddings
    total_paddings = []

    for idx in range(len(first_paddings)):
        total_paddings += [first_paddings[idx], last_paddings[idx]]

    total_paddings = torch.stack(total_paddings).long().tolist()

    # add padding offset to centers
    centers = centers + first_paddings[None]

    # pad tensor
    tensor = F.pad(tensor, total_paddings, **padding_kwargs)

    # pad additional tensor if given
    if additional_tensor is not None:
        additional_tensor = F.pad(additional_tensor, total_paddings, **padding_kwargs)

    # re-assign tensors
    for idx in range(out.size(0)):
        mins = (centers[idx] - max_range / 2.0).long()
        maxs = (centers[idx] + max_range / 2.0).long()
        if len(mins) == 2:
            out[idx] = tensor[idx][..., mins[0] : maxs[0], mins[1] : maxs[1]]
            if additional_tensor is not None:
                additional_out[idx] = additional_tensor[idx][..., mins[0] : maxs[0], mins[1] : maxs[1]]
        elif len(mins) == 3:
            out[idx] = tensor[idx][..., mins[0] : maxs[0], mins[1] : maxs[1], mins[2] : maxs[2]]
            if additional_tensor is not None:
                additional_out[idx] = additional_tensor[idx][
                    ..., mins[0] : maxs[0], mins[1] : maxs[1], mins[2] : maxs[2]
                ]
    # remove channel dim for mask_out if mask had no channel dim
    if additional_tensor is not None and additional_tensor.ndim == tensor.ndim - 1:
        additional_out = additional_out[:, 0]

    if additional_tensor is None:
        return out
    return out, additional_out


def crop_tensor_to_nonzero(
    tensor: torch.Tensor,
    additional_tensor: torch.Tensor | None = None,
    **padding_kwargs,
) -> tuple[torch.Tensor, torch.Tensor] | torch.Tensor:
    """
    crops tensor to non-zero and additional tensor to the same part of the
    tensor if given
    Args:
        tensor: the tensor to crop
        additional_tensor: an additional tensor to crop to the same region
            as :attr`tensor`
        **padding_kwargs: keyword arguments to controll the necessary padding
    Returns:
        torch.Tensor: the cropped tensor
        Optional[torch.Tensor]: the cropped additional tensor,
            only returned if passed
    """
    centers, ranges = extract_nonzero_bounding_box_from_tensor(tensor)

    return crop_tensor_to_bounding_box(
        tensor,
        centers=centers,
        ranges=ranges,
        additional_tensor=additional_tensor,
        **padding_kwargs,
    )


class CropOrPadPerImage(tio.transforms.CropOrPad):
    """Crops or pads every single image individually to the given shape."""

    def apply_transform(self, subject: tio.data.Subject) -> tio.data.Subject:
        """
        Args:
            subject: the subject to apply the transform to

        Returns:
            the subject with the transformed data

        """
        for k, v in subject.get_images_dict(intensity_only=False, include=self.include, exclude=self.exclude).items():
            part_sub = type(subject)({k: v})
            subject[k] = super().apply_transform(part_sub)[k]
        return subject


# TODO: Add Transform to cut only one side of the image and flip it.



# Copyright Justus Schock.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import gc
import importlib
import json
import os
from abc import ABCMeta, abstractmethod
from collections import Counter, namedtuple
from copy import deepcopy
from functools import partial
from itertools import chain
from operator import attrgetter, itemgetter
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union
from numbers import Number

import torch
import torchio as tio
import tqdm
from tqdm.contrib.concurrent import process_map

from mdlu.data.modality import ImageModality
from mdlu.utils import PyTorchJsonDecoder, PyTorchJsonEncoder

ImageStats = namedtuple("ImageStats", ("spacing", "spatial_shape", "unique_intensities", "intensity_counts", "num_channels"))

__all__ = ["AbstractDataset", "AbstractDiscreteLabelDataset"]


# TODO: Add possibility to also add intensity mean and std manually
class AbstractDataset(tio.data.SubjectsDataset, metaclass=ABCMeta):
    """Abstract Dataset class defining the interface for all datasets. For usage, subclass this class and implement
    at least ``parse_subjects``.

    The workflow within the dataset is as follows:

      ┌─────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┐
      │``__init__``            ┌──────────┐                                                              ┌────────────────┐             │
      │                        │First Run │                                                              │ Subsequent Run │             │
      │                        └──────────┘                                                              └────────────────┘             │
      │                             │                                                                            │                      │
      │                             │                                                                   ┌────────┴──────────┐           │
      │                             ▼                                                                   ▼                   ▼           │
      │                      ┌──────────────┐                                                    ┌──────────────┐  ┌──────────────────┐ │
      │                      │Parse Subjects│                                                    │Parse Subjects│  │Parse Preprocessed│ │
      │                      │              │                                                    │              │  │     Subjects     │ │
      │                      └──────────────┘                                                    └──────────────┘  └──────────────────┘ │
      │                              ┌────────────────────────Not Equal──────────────┐                   │                   │          │
      │                              │                                               │                   └─────────┬─────────┘          │
      │                              ▼                                               │                             ▼                    │
      │ ┌─────────────────────────────────────────────────────────┐                  │                 ┌───────────────────────┐        │
      │ │                    For each Sample:                     │                  │                 │Check whether Number of│        │
      │ │  ┌─────────────┐   (Optionally Parallelized)            │                  └─────────────────│   Subjects is equal   │        │
      │ │  │ Load Sample │──┬─────────────────────┐               │                                    └───────────────────────┘        │
      │ │  └─────────────┘  │                     │               │                                                │                    │
      │ │                   │                     │               │                                             Equal                   │
      │ │              ┌────┘                     │               │                                                │                    │
      │ │              │                          │               │                                                │                    │
      │ │              ▼                          ▼               │                                                ▼                    │
      │ │  ┌──────────────────────┐   ┌──────────────────────┐    │                                    ┌───────────────────────┐        │
      │ │  │    Extract image     │   │    Extract Label     │    │                                    │ Load Image And Label  │        │
      │ │  │      Statistics      │   │      Statistics      │    │                                    │      Statistics       │        │
      │ │  └──────────────────────┘   └──────────────────────┘    │                                    └───────────────────────┘        │
      │ │              │                          │               │                                                │                    │
      │ └──────────────┼──────────────────────────┼───────────────┘                                                │                    │
      │                ▼                          ▼                                                                │                    │
      │    ┌──────────────────────┐   ┌──────────────────────┐                                                     │                    │
      │    │ Aggregate All Image  │   │ Aggregate All Label  │                                                     │                    │
      │    │      Statistics      │   │      Statistics      │                                                     │                    │
      │    │                      │   │                      │                                                     │                    │
      │    └──────────────────────┘   └──────────────────────┘                                                     │                    │
      │                │                          │                                                                │                    │
      │                │                          │                                           ┌────────────────────┤                    │
      │                └────────────┬─────────────┘                                           │                    │                    │
      │                             │                                                         │                    │                    │
      |                             ▼                      ┌─────────────────────────┐        │                    │                    │
      │               ┌──────────────────────────┐         │    Save Preprocessed    │        │                    │                    │
      │               │  Preprocess All Samples  │────────▶│         Samples         │        │                    │                    │
      │               └──────────────────────────┘         │(Optionally Parallelized)│        │                    │                    │
      │                             │                      └─────────────────────────┘        │                    │                    │
      │                             ▼                                   │                     │                    │                    │
      │               ┌──────────────────────────┐                      │                     ▼                    │                    │
      │               │   Save Image And Label   │                      │ ┌──────────────────────────────────────┐ │                    │
      │               │        Statistics        │                      │ │                                      │ │                    │
      │               └──────────────────────────┘                      │ │ Instantiate Augmentation Transforms  │ │                    │
      │                             │                         ┌─────────┼▶│                                      │ │                    │
      │                             │                         │         │ │                                      │ │                    │
      │                             └─────────────────────────┘         │ └──────────────────────────────────────┘ │                    │
      │                                                                 └─────────────────────┬────────────────────┤                    │
      │                                                                                       │                    │                    │
      │                                                                                       │                    │                    │
      │                                                                                       │                    │                    │
      │                                                                                       │                    │                    │
      └───────────────────────────────────────────────────────────────────────────────────────┼────────────────────┼────────────────────┘
                                                                                              │                    │
                                                                                              │                    │
                                ┌─────────────────────────────────────────────────────────────┼────────────────────▼───────────────────┐
    ┌───────────────────┐       │``__getitem__``                                              │ ┌────────────────────────────────────┐ │
    │                   │       │                                                             │ │                                    │ │
    │Dataloader queries │       │                  Fetch Item                                 │ │List of Preprocessed Samples On Disk│ │
    │  Dataset Sample   │───────┼──────────────────From List──────────────────────────────────┼▶│        (Not Cached in RAM)         │ │
    │                   │       │                                                             │ │                                    │ │
    └───────────────────┘       │                                                             │ └────────────────────────────────────┘ │
            ▲                   │                                                             │                    │                   │
            │                   │                                                             │                    │                   │
            │                   │                                                             │                    │                   │
            │                   │                                                             │                    │                   │
            │                   │                                                             │             Load   │                   │
            │                   │                                                             ├─────────Preprocessed                   │
            │                   │                                                             │            Sample                      │
            │                   │                                                             │                                        │
            │                   │                                                             │                                        │
            │                   │                                                             │                                        │
            │                   │                                                             ▼                                        │
            │                   │                         ┌──────────────────────────────────────────────────────────────────────┐     │
            │                   │      Return             │                                                                      │     │
            └───────────────────┼────Augmented ───────────│                          Apply Augmentation                          │     │
                                │      Sample             │                                                                      │     │
                                │                         └──────────────────────────────────────────────────────────────────────┘     │
                                │                                                                                                      │
                                │                                                                                                      │
                                │                                                                                                      │
                                │                                                                                                      │
                                └──────────────────────────────────────────────────────────────────────────────────────────────────────┘

    For extracting custom image and label statistics, you can overwrite the following ``staticmethods`` :

    - ``get_single_image_stats``
    - ``get_single_label_stats``
    - ``aggregate_image_stats``
    - ``aggregate_label_stats``

    And add your extracted statistics to the ``image_stat_attr_keys`` and ``label_stat_attr_keys`` tuples.

    Per default the following statistics are extracted:

    - image spacings
    - spatial image shape
    - image intensity occurence counts

    and based on that the following attributes are calculated:

    - :attr:`median_spacing` : the median spacing of all images
    - :attr:`computed_target_spacing` : the target spacing computed by the dataset
    - :attr:`target_spacing` : the externally passed target spacing if provided else the computed one
    - :attr:`sizes_after_resampling` : the sizes of the images after resampling to target spacing
    - :attr:`median_size_after_resampling` : the median size of all images after resampling to target spacing
    - :attr:`target_size` : the externally passed target size if provided else the computed one
    - :attr:`mean_intensity_value` : the mean intensity value of all images
    - :attr:`std_intensity_value` : the standard deviation of all images
    - :attr:`max_size_after_resampling` : the maximum size of all images after resampling to target spacing
    - :attr:`min_size_after_resampling` : the minimum size of all images after resampling to target spacing
    - :attr:`default_preprocessing` : the default preprocessing applied to all images and labels
    - :attr:`default_augmentation` : the default augmentation applied to all images and labels

    Per default, no label statistics are extracted, as the labels are highly task-dependent.
    """

    extension_mapping = {
        tio.constants.INTENSITY: ".nii.gz",
        tio.constants.LABEL: ".nii.gz",
    }
    restore_mapping = {
        tio.constants.INTENSITY: tio.data.ScalarImage,
        tio.constants.LABEL: tio.data.LabelMap,
    }
    image_stat_attr_keys: tuple[str, ...] = (
        "spacings",
        "spatial_shapes",
        "intensity_counts",
        "num_channels_all_images",
    )
    label_stat_attr_keys: tuple[str, ...] = ()

    pre_stats_trafo = tio.transforms.ToCanonical()

    spacings: torch.Tensor
    spatial_shapes: torch.Tensor
    intensity_counts: Counter
    num_channels: torch.Tensor

    def __init__(
        self,
        *paths: Path | str,
        preprocessed_path: Path | str | None,
        image_modality: ImageModality | int | str,
        image_stat_key: str | None = None,
        label_stat_key: str | None = None,
        preprocessing: (tio.transforms.Transform | Callable[[tio.data.Subject], tio.data.Subject] | None) = "default",
        augmentation: tio.transforms.Transform | Callable[[tio.data.Subject], tio.data.Subject] | None = None,
        statistic_collection_nonzero: bool = False,
        num_stat_collection_procs: int = 0,
        num_save_procs: int = 0,
        anisotropy_threshold: int = 3,
        target_spacing: Sequence[float] | torch.Tensor | None = None,
        target_size: Sequence[int] | torch.Tensor | None = None,
    ):

        """

        Args:
            paths: Paths to the data (images, labels and everything else). Just used to pass 1:1 to ``parse_subjects``.
            preprocessed_path: Path to the preprocessed data. If not None, will attempt to load the preprocessed data
                if the path exists and will sotre the preprocessed data there if the path does not yet exist.
            image_modality: The modality of the image. Can be either an instance of :class:`ImageModality`, an integer or string from the following list:
                - ``PHOTOGRAPH`` | 0
                - ``XRAY``       | 1
                - ``MR``         | 2
                - ``CT``         | 3
            image_stat_key: The key of the image within the subject to extract the statistics from. If None, will use the first image.
            label_stat_key: The key of the label within the subject to extract the statistics from. If None, won't extract any statistics.
            preprocessing: The preprocessing to apply to the subjects. If None, will not apply any preprocessing. If "default", will apply the :attr`default_preprocessing`.
            augmentation: The augmentation to apply to the subjects. If None, will not apply any augmentation. If "default", will apply the :attr`default_augmentation`.
            statistic_collection_nonzero: If True, will crop to nonzero before collecting statistics.
            num_stat_collection_procs: The number of processes to use for collecting the statistics.
                If 0, will only use the main process. If 1, will also just use one process (but a spawned one), if >1, will use multiple processes.
                Using 0 and 1 will be similarly fast, but using 0 won't do any pickling. Switch between 0 and 1 for debugging purposes regarding pickling.
            num_save_procs: The number of processes to use for saving the data.
                If 0, will only use the main process. If 1, will also just use one process (but a spawned one), if >1, will use multiple processes.
            anisotropy_threshold: If largest spacing > (anisotropy_threshold * second largest spacing), an image is considered to be anisotropic.
            target_spacing: The target spacing to use for resampling. If None, will resample to the computed target spacing.
            target_size: The target size to use for resampling. If None, will crop or pad to the computed target size.
        """
        # need to do this early on as torch hacks some custom things (like __getattr__)
        torch.utils.data.Dataset.__init__(self)
        parsed_subjects = self.parse_subjects(*paths)
        if isinstance(image_modality, str) and not isinstance(image_modality, ImageModality):
            image_modality = ImageModality.from_str(image_modality)

        self.image_modality = image_modality
        self.image_stats_key = image_stat_key
        self.label_stats_key = label_stat_key

        self.anisotropy_threshold = anisotropy_threshold

        if not isinstance(target_spacing, torch.Tensor) and target_spacing is not None:
            target_spacing = torch.tensor(target_spacing)
        self._target_spacing = target_spacing

        if not isinstance(target_size, torch.Tensor) and target_size is not None:
            target_size = torch.tensor(target_size, dtype=torch.long)
        self._target_size = target_size

        preprocessed_parsed_subjects: list[tio.data.Subject] | tuple | None = None
        if preprocessed_path is not None and os.path.exists(preprocessed_path):
            try:
                preprocessed_parsed_subjects, dataset_stats = self.restore_preprocessed(preprocessed_path)
            except Exception as e:
                # raise Warning for exception
                preprocessed_parsed_subjects, dataset_stats = (), None

            # Incorrect Saving/Loading -> Trigger warning
            if len(preprocessed_parsed_subjects) != len(parsed_subjects):
                preprocessed_parsed_subjects, dataset_stats = None, None
        else:
            preprocessed_parsed_subjects, dataset_stats = None, None

        if dataset_stats is None:
            dataset_stats = self.collect_dataset_stats(
                *parsed_subjects,
                image_stat_key=image_stat_key,
                label_stat_key=label_stat_key,
                crop_to_nonzero=statistic_collection_nonzero,
                num_procs=num_stat_collection_procs,
            )

        self.set_image_stat_attributes(dataset_stats["image"])
        self.set_label_stat_attributes(dataset_stats["label"])

        transforms = []

        preprocessing_trafo = self.get_preprocessing_transforms(preprocessing)
        if preprocessed_parsed_subjects is None and preprocessed_path is not None:
            self.save_preprocessed(
                *parsed_subjects,
                save_path=preprocessed_path,
                preprocessing_trafo=preprocessing_trafo,
                num_procs=num_save_procs,
            )
            preprocessed_parsed_subjects, _ = self.restore_preprocessed(preprocessed_path)

        elif preprocessed_path is not None and preprocessing_trafo is not None:
            transforms.append(preprocessing_trafo)

        augmentations = self.get_augmentation_transforms(augmentation)

        if augmentations is not None:
            transforms.append(augmentations)

        super().__init__(
            preprocessed_parsed_subjects or parsed_subjects,
            transform=tio.transforms.Compose(transforms),
            load_getitem=True,
        )

    @abstractmethod
    def parse_subjects(self, *paths: Path | str) -> Sequence[tio.data.Subject]:
        """Parses the subjects from the given paths. To be implemented in subclasses by the user.

        Args:
            paths: Paths to the data (images, labels and everything else).
        Returns:
            A sequence of subjects (ideally not yet loaded).
        """
        pass

    def set_image_stat_attributes(self, image_stats: dict[str, torch.Tensor | Counter]):
        """Sets the image statistics attributes.

        Args:
            image_stats: The image statistics to set as attributes.
        """
        for name in self.image_stat_attr_keys:
            setattr(self, name, image_stats[name])

    def set_label_stat_attributes(self, label_stats: dict[str, Any]):
        """Sets the label statistics attributes.

        Args:
            label_stats: The label statistics to set as attributes.
        """
        for name in self.label_stat_attr_keys:
            setattr(self, name, label_stats[name])

    @staticmethod
    def get_single_image_stats(image: tio.data.Image) -> Any | ImageStats:
        """Gets the image statistics for a single image.

        Args:
            image: The image to get the statistics for.
        Returns:
            The image statistics (spacing, spatial_shape, unique values, occurence counts).
        """
        uniques, counts = image.tensor[image.tensor > image.tensor.min()].unique(return_counts=True)
        uniques, counts = uniques.tolist(), counts.tolist()

        return ImageStats(image.spacing, image.spatial_shape, uniques, counts, image.tensor.size(0))

    @staticmethod
    def get_single_label_stats(label: Any, *args: Any, **kwargs: Any) -> Any:
        """Gets the label statistics for a single label.

        Args:
            label: The label to get the statistics for.

        Returns:
            The label statistics (None per default)
        """
        pass

    @staticmethod
    def aggregate_image_stats(*image_stats: Any | ImageStats) -> dict[str, Any | torch.Tensor]:
        """Aggregates the image statistics.

        Args:
            image_stats: The image statistics to aggregate.

        Returns:
            Dictionary with aggregated image statistics.
        """
        intensity_values: Counter[float] = Counter()

        spacings = tuple(map(attrgetter("spacing"), image_stats))
        shapes = tuple(map(attrgetter("spatial_shape"), image_stats))
        unique_intensities = map(attrgetter("unique_intensities"), image_stats)
        intensity_counts = map(attrgetter("intensity_counts"), image_stats)
        num_channels = tuple(map(attrgetter("num_channels"), image_stats))

        for u, c in zip(unique_intensities, intensity_counts):
            intensity_values.update(dict(zip(u, c)))

        return {
            "spacings": torch.tensor(spacings, dtype=torch.float),
            "spatial_shapes": torch.tensor(shapes, dtype=torch.long),
            "intensity_counts": intensity_values,
            "num_channels_all_images": torch.tensor(num_channels, dtype=torch.long)
        }

    @staticmethod
    def aggregate_label_stats(*stats: Any) -> dict[str, Any] | None:
        """Aggregates the label statistics.

        Args:
            stats: The label statistics to aggregate.

        Returns:
            Dictionary with aggregated label statistics.
        """
        pass

    def collect_stats_single_subject(
        self,
        subject: tio.data.Subject,
        image_stat_key: str | None,
        label_stat_key: str | None,
        crop_to_nonzero: bool = False,
    ) -> tuple[ImageStats | Any, Any | None]:
        """Collects the statistics for a single subject.

        Args:
            subject: The subject to collect the statistics for.
            image_stat_key: The key to use for the image statistics.
            label_stat_key: The key to use for the label statistics.
            crop_to_nonzero: Whether to crop the image to non-zero values.
        Returns:
            The image statistics and the label statistics.
        """

        # do all the copying here to avoid loading the acutal subject.
        # loading a copy is fine as this will be garbage collected at the end of the function
        # loading the original subject which is also referenced somewhere else would lead to a memory leak
        subject_copy = deepcopy(subject)

        subject = self.pre_stats_trafo(subject_copy)

        if crop_to_nonzero:
            from mdlu.transforms import CropToNonZero

            trafo = CropToNonZero()
            subject = trafo(subject)

        # TODO: Add Warnings if keys not present
        if image_stat_key:
            image = subject_copy.get(image_stat_key, subject_copy.get_first_image())
        else:
            image = subject_copy.get_first_image()

        if label_stat_key:
            label = subject_copy.get(label_stat_key, None)
        else:
            label = None

        stats = deepcopy((self.get_single_image_stats(image), self.get_single_label_stats(label)))
        del subject_copy
        del image
        del label
        gc.collect()

        return stats

    def collect_dataset_stats(
        self,
        *subjects: tio.data.Subject,
        image_stat_key: str | None,
        label_stat_key: str | None,
        crop_to_nonzero: bool,
        num_procs: int,
    ) -> dict[str, dict[str, Any | None]]:
        """Collects the statistics for the whole dataset.

        Args:
            subjects: The subjects to collect the statistics for.
            image_stat_key: The key to use for the image statistics.
            label_stat_key: The key to use for the label statistics.
            crop_to_nonzero: Whether to crop the image to non-zero values.
            num_procs: The number of processes to use.

        Returns:
            The collected statistics.
        """
        func = partial(
            self.collect_stats_single_subject,
            image_stat_key=image_stat_key,
            label_stat_key=label_stat_key,
            crop_to_nonzero=crop_to_nonzero,
        )

        desc = "Retrieving Subject Statistics"
        if num_procs == 0:
            results = list(map(func, tqdm.tqdm(subjects, desc=desc)))
        else:
            results = process_map(func, subjects, max_workers=num_procs, desc=desc)

        return {
            "image": self.aggregate_image_stats(
                *map(itemgetter(0), results),
            ),
            "label": self.aggregate_label_stats(*map(itemgetter(1), results)),  # type: ignore
        }

    def get_preprocessing_transforms(self, preprocessing) -> tio.transforms.Transform | None:
        """Gets the preprocessing transforms.

        Args:
            preprocessing: The preprocessing to get the transforms for.

        Returns:
            The preprocessing transforms.

        Raises:
            ValueError: If the preprocessing is not supported.
        """
        if preprocessing is None or (isinstance(preprocessing, str) and preprocessing.lower() == "none"):
            return None
        elif callable(preprocessing):
            return preprocessing

        elif isinstance(preprocessing, str) and preprocessing == "default":
            return self.default_preprocessing

        else:
            raise ValueError(f"Invalid Preprocessing: {preprocessing}")

    def get_augmentation_transforms(self, augmentation) -> tio.transforms.Transform | None:
        """Gets the augmentation transforms.

        Args:
            augmentation: The augmentation to get the transforms for.

        Returns:
            The augmentation transforms.

        Raises:
            ValueError: If the augmentation is not supported.
        """
        if augmentation is None or (isinstance(augmentation, str) and augmentation.lower() == "none"):
            return None
        elif callable(augmentation):
            return augmentation

        elif isinstance(augmentation, str) and augmentation == "default":
            return self.default_augmentation

        else:
            raise ValueError(f"Invalid Augmentation: {augmentation}")

    def save_single_preprocessed_subject(
        self,
        idx: int,
        subject: tio.data.Subject,
        save_path: str | Path,
        preprocessing_trafo: tio.transform.Transform | Callable[[tio.data.Subject], tio.data.Subject] | None,
        total_num_subjects: int,
    ) -> None:
        """Preprocesses and saves a single preprocessed subject to a json file for metadata, an image and a label
        file,

        Args:
            idx: The index of the subject.
            subject: The subject to save.
            save_path: The path to save the subject to.
            preprocessing_trafo: The preprocessing transform to use.
            total_num_subjects: The total number of subjects.
        """
        # do all the copying here to avoid loading the acutal subject.
        # loading a copy is fine as this will be garbage collected at the end of the function
        # loading the original subject which is also referenced somewhere else would lead to a memory leak
        subject_copy = deepcopy(subject)
        if preprocessing_trafo is not None:
            subject_copy = preprocessing_trafo(subject_copy)

        save_path = os.path.join(save_path, str(idx).zfill(len(str(total_num_subjects))))
        os.makedirs(save_path, exist_ok=True)

        tio_images = {}
        to_dump = {}
        for k, v in subject_copy.items():
            if isinstance(v, tio.data.Image):
                tio_images[k] = v.type
                v.save(os.path.join(save_path, k + self.extension_mapping[v.type]))
            else:
                to_dump[k] = v

        to_dump["TORCHIO_SUBJECT_CLASS"] = ".".join((type(subject).__module__, type(subject).__qualname__))
        to_dump["TORCHIO_IMAGE_TYPES"] = tio_images

        with open(os.path.join(save_path, "subject.json"), "w") as f:
            json.dump(to_dump, f, indent=4, sort_keys=True, cls=PyTorchJsonEncoder)

    def _wrapped_save_single_preprocessed_subject(
        self,
        args: tuple[int, tio.data.Subject],
        save_path: str | Path,
        preprocessing_trafo: tio.transform.Transform | Callable[[tio.data.Subject], tio.data.Subject] | None,
        total_num_subjects: int,
    ) -> None:
        """Wrapper for the save_single_preprocessed_subject function to enable its usage with multiprocessing.

        Args:
            args: The arguments forward to the function (usually index and Subject).
            save_path: The path to save the subject to.
            preprocessing_trafo: The preprocessing transform to use.
            total_num_subjects: The total number of subjects.
        """
        self.save_single_preprocessed_subject(
            *args, save_path=save_path, preprocessing_trafo=preprocessing_trafo, total_num_subjects=total_num_subjects
        )
        return self.save_single_preprocessed_subject(
            *args,
            save_path=save_path,
            preprocessing_trafo=preprocessing_trafo,
            total_num_subjects=total_num_subjects,
        )

    def image_state_dict(self) -> dict[str, Any]:
        """Returns the state dict of the image statistics across the whole dataset.

        Returns:
            The state dict of the image statistics.
        """
        return {k: getattr(self, k) for k in self.image_stat_attr_keys}

    def label_state_dict(self) -> dict[str, Any]:
        """Returns the state dict of the label statistics across the whole dataset.

        Returns:
            The state dict of the label statistics.
        """

        return {k: getattr(self, k) for k in self.label_stat_attr_keys}

    def state_dict(self) -> dict[str, dict[str, Any]]:
        """Returns the combined state dict for images and labels of the statistics across the whole dataset.

        Returns:
            The state dict of the statistics.
        """
        return {"image": self.image_state_dict(), "label": self.label_state_dict()}

    def save_preprocessed(self, *subjects, save_path, preprocessing_trafo, num_procs) -> None:
        """Preprocesses and saves all subjects in the dataset to a json file for metadata, an image and a label
        file each. Also saves a json file containing the state_dict of the dataset.

        Args:
            subjects: The subjects to save.
            save_path: The path to save the subjects to.
            preprocessing_trafo: The preprocessing transform to use.
            num_procs: The number of processes to use for multiprocessing.
        """
        func = partial(
            self._wrapped_save_single_preprocessed_subject,
            save_path=save_path,
            preprocessing_trafo=preprocessing_trafo,
            total_num_subjects=len(subjects),
        )

        desc = "Saving Preprocessed Subjects"
        if num_procs == 0:
            list(map(func, enumerate(tqdm.tqdm(subjects, desc=desc))))
        else:
            process_map(
                func,
                enumerate(subjects),
                max_workers=num_procs,
                desc=desc,
                total=len(subjects),
            )

        stat_dict = self.state_dict()

        with open(os.path.join(save_path, "dataset.json"), "w") as f:
            json.dump(
                stat_dict,
                f,
                indent=4,
                sort_keys=True,
                cls=PyTorchJsonEncoder,
            )

    def restore_preprocessed(self, preprocessed_path) -> tuple[list[tio.data.Subject], dict[str, Any]]:
        """Restores the preprocessed subjects from a json file for metadata, an image and a label file each. Also
        restores the state_dict of the dataset.

        Args:
            preprocessed_path: The path to the preprocessed subjects.

        Returns:
            The subjects and the state dict of the dataset.
        """
        with open(os.path.join(preprocessed_path, "dataset.json")) as f:
            dset_meta = json.load(f, cls=PyTorchJsonDecoder)

        subjects = []
        subs = sorted(
            (
                os.path.join(preprocessed_path, x)
                for x in os.listdir(preprocessed_path)
                if os.path.isdir(os.path.join(preprocessed_path, x))
            ),
            key=lambda _x: int(os.path.split(_x)[1]),
        )

        for sub in subs:
            # load information about subject
            with open(os.path.join(sub, "subject.json")) as f:
                subject_meta = json.load(f, cls=PyTorchJsonDecoder)

            subject_cls_path = subject_meta.pop("TORCHIO_SUBJECT_CLASS", "")

            if subject_cls_path:

                cls_module_path, cls_name = subject_cls_path.rsplit(".", 1)
                cls_module = importlib.import_module(cls_module_path)

                if cls_module is not None:
                    subject_cls = getattr(cls_module, cls_name, None)
                else:
                    subject_cls = None
            else:
                subject_cls = None

            subject_cls = subject_cls or tio.data.Subject

            images = subject_meta.pop("TORCHIO_IMAGE_TYPES", {})
            for k, v in images.items():
                subject_meta[k] = self.restore_mapping[v](os.path.join(sub, k + self.extension_mapping[v]))

            subjects.append(subject_cls(subject_meta))

        return subjects, dset_meta

    @property
    def median_spacing(self) -> torch.Tensor:
        """Computes the median spacing of the dataset.

        Returns:
            The median spacing of the dataset.
        """
        return torch.median(self.spacings, 0).values

    @property
    def computed_target_spacing(self) -> torch.Tensor:
        """Computes the target spacing of the dataset under considereation of possible anisotrophies and the
        expected size after resampling to median spacing.

        Returns:
            The computed target spacing of the dataset.
        """
        target_spacing = self.median_spacing
        target_size = self.median_size_after_resampling
        # we need to identify datasets for which a different target spacing could be beneficial. These datasets have
        # the following properties:
        # - one axis which much lower resolution than the others
        # - the lowres axis has much less voxels than the others
        # - (the size in mm of the lowres axis is also reduced)
        worst_spacing_axis = torch.argmax(target_spacing)
        other_axes = [i for i in range(len(target_spacing)) if i != worst_spacing_axis]
        other_spacings = [target_spacing[i] for i in other_axes]
        other_sizes = [target_size[i] for i in other_axes]

        has_aniso_spacing = target_spacing[worst_spacing_axis] > (self.anisotropy_threshold * max(other_spacings))
        has_aniso_voxels = target_size[worst_spacing_axis] * self.anisotropy_threshold < min(other_sizes)

        if has_aniso_spacing and has_aniso_voxels:
            spacings_of_that_axis = self.spacings[:, worst_spacing_axis]
            target_spacing_of_that_axis = torch.quantile(spacings_of_that_axis, 0.1)
            # don't let the spacing of that axis get higher than the other axes
            if target_spacing_of_that_axis < max(other_spacings):
                target_spacing_of_that_axis = max(max(other_spacings), target_spacing_of_that_axis) + 1e-5
            target_spacing[worst_spacing_axis] = target_spacing_of_that_axis

        return target_spacing

    @property
    def target_spacing(self) -> torch.Tensor:
        """Returns the externally given target spacing if provided and falls back to computing it.

        Returns:
            The target spacing for the dataset
        """
        return self._target_spacing or self.computed_target_spacing

    @property
    def num_channels(self) -> int:
        if self.num_channels_all_images.unique(return_counts=False) == 1:
            return self.num_channels_all_images[0].item()
        raise RuntimeError(
            f"""All images need to have the same number of channels, 
            but got images with the following numbers of channels: {self.num_channels_all_images}"""
        )

    @property
    def sizes_after_resampling(self) -> torch.Tensor:
        """Computes the image sizes of the dataset after resampling to the median spacing.

        Returns:
            The image sizes after resampling
        """
        return (self.spatial_shapes.float() / self.spacings * self.median_spacing).round().long()

    @property
    def median_size_after_resampling(self) -> torch.Tensor:
        """Computes the median image size of the dataset after resampling to the median spacing.

        Returns:
            The median image size after resampling
        """
        return torch.median(self.sizes_after_resampling, 0).values

    @property
    def target_size(self) -> torch.Tensor:
        """Returns the externally given target size if provided and falls back to computing it as the median size
        after resampling.

        Returns:
            The target size for the dataset
        """
        return self._target_size or self.median_size_after_resampling

    @property
    def mean_intensity_value(self) -> torch.Tensor:
        """Computes the mean image intensity value of the dataset.

        Returns:
            The mean intensity value of the dataset.
        """
        return torch.tensor(
            sum(float(k) * float(v) for k, v in self.intensity_counts.items()) / sum(self.intensity_counts.values())
        )

    @property
    def std_intensity_value(self) -> torch.Tensor:
        """Computes the standard deviation of the image intensity values of the dataset.

        Returns:
            The standard deviation of the image intensity values.
        """

        n = sum(self.intensity_counts.values())
        # normal intensity calculation without allocating all the occurences to save memory
        return (
            (
                sum(float(k) * float(v) ** 2 for k, v in self.intensity_counts.items())
                - n * self.mean_intensity_value ** 2
            )
            / (n - 1)
        ).sqrt()

    @property
    def max_size_after_resampling(self) -> torch.Tensor:
        """Computes the maximum image size of the dataset after resampling to the median spacing.

        Returns:
            The maximum image size after resampling.
        """
        return torch.max(self._sizes_after_resampling, 0).values

    @property
    def min_size_after_resampling(self) -> torch.Tensor:
        """Computes the minimum image size of the dataset after resampling to the median spacing.

        Returns:
            The minimum image size after resampling.
        """
        return torch.min(self._sizes_after_resampling, 0).values

    @property
    def default_preprocessing(self) -> Callable[[tio.data.Subject], tio.data.Subject] | tio.transforms.Transform | None:
        """Returns the default preprocessing for the dataset.

        Returns:
            The default preprocessing for the dataset.
        """
        from mdlu.transforms import DefaultPreprocessing

        return DefaultPreprocessing(
            target_spacing=self.target_spacing.tolist(),
            target_size=self.target_size.tolist(),
            modality=self.image_modality,
            dataset_intensity_mean=self.mean_intensity_value,
            dataset_intensity_std=self.std_intensity_value,
            affine_key=self.image_stats_key,
        )

    @property
    def default_augmentation(self) -> Callable[[tio.data.Subject], tio.data.Subject] | tio.transforms.Transform | None:
        """Returns the default augmentation for the dataset.

        Returns:
            The default augmentation for the dataset.
        """
        from mdlu.transforms import DefaultAugmentation

        return DefaultAugmentation(
            self.image_modality,
            include_deformation=False,
        )


class AbstractDiscreteLabelDataset(AbstractDataset):
    """Abstract dataset for a task with discrete labels (like semantic segmentation or classification).

    Additionally to the image statistics it also stores the class values as label statistics and derives the following attributes:

    - :attr:`consecutive_class_mapping`: A mapping from consecutive class values to the original class values.
    - :attr:`inverse_class_mapping`: A mapping from original class values to consecutive class values.
    - :attr`num_classes`: The number of classes in the dataset.

    For more information see :class:`AbstractDataset`.
    """

    class_values: torch.Tensor
    label_stat_attr_keys = ("class_values", "spatial_label_shapes", *AbstractDataset.label_stat_attr_keys)

    @staticmethod
    def get_single_label_stats(label: tio.data.Image | torch.Tensor | Number, *args: Any, **kwargs: Any) -> dict[str, torch.Tensor | Any] | Any:
        """Computes the label statistics for a single label.

        Args:
            label: The label to compute the statistics for.

        Returns:
            The label statistics for the label.
        """
        if isinstance(label, tio.data.Image):
            label_tensor = label.tensor
        elif isinstance(label, torch.Tensor):
            label_tensor = label
        elif isinstance(label, Number):
            label_tensor = torch.tensor(label).view(1,)
        else:
            raise TypeError(f'label has to be either torchio Image, torch.Tensor or Number, got {type(label)}.')
        return {"class_values": label_tensor.unique().tolist(), 'spatial_shape': label_tensor.shape[1:]}

    @staticmethod
    def aggregate_label_stats(*label_stats: Any) -> dict[str, torch.Tensor | Any] | None:
        """Aggregates the label statistics of multiple labels.

        Args:
            label_stats: The label statistics of multiple labels.

        Returns:
            The aggregated label statistics.
        """
        return {
            "class_values": torch.tensor(sorted(set(chain.from_iterable(map(itemgetter("class_values"), label_stats))))),
            "spatial_label_shapes": torch.tensor(tuple(map(attrgetter("spatial_shape"), image_stats)), dtype=torch.long)
        }

    @property
    def target_label_size(self) -> torch.Tensor:
        if torch.allclose(self.spatial_label_shapes, self.spatial_shapes):
            return self.target_size
        
        median = spatial_shapes.median(0)
        if torch.isnan(median).any():
            return torch.tensor([1])
        return median

    @property
    def consecutive_class_mapping(self) -> dict[int, int]:
        """Returns the mapping from consecutive class values to the original class values.

        Returns:
            The mapping from consecutive class values to the original class values.
        """
        return dict(enumerate(self.class_values.tolist()))

    @property
    def inverse_class_mapping(self) -> dict[int, int]:
        """Returns the mapping from original class values to consecutive class values.

        Returns:
            The mapping from original class values to consecutive class values.
        """
        return {v: k for k, v in self.consecutive_class_mapping.items()}

    @property
    def num_classes(self) -> int:
        """Returns the number of classes in the dataset.

        Returns:
            The number of classes in the dataset.
        """
        return self.class_values.numel()

    @property
    def default_preprocessing(self) -> Callable[[tio.data.Subject], tio.data.Subject] | tio.transforms.Transform | None:
        """Returns the default preprocessing for the dataset.

        Returns:
            The default preprocessing for the dataset.
        """
        from mdlu.transforms import DefaultPreprocessing

        return tio.transforms.Compose(
            [
                tio.transforms.RemapLabels(self.inverse_class_mapping),
                DefaultPreprocessing(
                    target_spacing=self.target_spacing.tolist(),
                    target_size=self.median_size_after_resampling.long().tolist(),
                    modality=self.image_modality,
                    dataset_intensity_mean=self.mean_intensity_value,
                    dataset_intensity_std=self.std_intensity_value,
                    affine_key=self.image_stats_key,
                    num_classes=self.num_classes,
                ),
            ]
        )



# Copyright Justus Schock.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
from contextlib import contextmanager
import json
from typing import Any, Mapping, Generator

import torch
from pytorch_lightning.utilities.apply_func import apply_to_collection

__all__ = [
    "PyTorchJsonEncoder",
    "PyTorchJsonDecoder",
]


class PyTorchJsonEncoder(json.JSONEncoder):
    """An encoder for PyTorch Tensors."""

    def default(self, o: Any) -> Any:
        """Encode a PyTorch Tensor."""
        if isinstance(o, torch.Tensor):
            o = {"type": "torch.Tensor", "content": o.tolist()}
        return o


class PyTorchJsonDecoder(json.JSONDecoder):
    """A decoder for serialized PyTorch Tensors."""

    def to_tensor(self, d: Mapping[str, Any]) -> torch.Tensor:
        """convert mappings to tensors if applicable."""
        if isinstance(d, Mapping) and "type" in d and d["type"] == "torch.Tensor" and "content" in d:
            return torch.tensor(d["content"])
        return {k: apply_to_collection(v, Mapping, self.to_tensor) for k, v in d.items()}

    def decode(self, s: str, *args, **kwargs) -> Any:
        decoded = super().decode(s, *args, **kwargs)
        decoded = apply_to_collection(decoded, Mapping, self.to_tensor)
        return decoded


# TODO: Find why this is not suppressing all PL Trainer output
@contextmanager
def suppress_stdout() -> Generator[None, None, None]:
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        old_std_err = sys.stderr
        sys.stderr = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_std_err



# Copyright Justus Schock.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from pytorch_lightning.utilities.enums import LightningEnum

__all__ = ["ImageModality"]


class ImageModality(LightningEnum):
    """Enum class for image modalities."""

    PHOTOGRAPH = 0
    XRAY = 1
    MR = 2
    CT = 3

    @staticmethod
    def get_dimensionality(modality: int | ImageModality) -> int:
        """Get the dimensionality of the image modality.

        Args:
            modality: The image modality.

        Returns:
            The dimensionality of the image modality (Usually 2D or 3D).
        """
        return 2 + int(int(modality) > 1)

    @classmethod
    def from_str(cls, value: str) -> ImageModality:
        """Get the image modality from a string.

        Args:
            value: The string value.

        Returns:
            The image modality.

        Raises:
            ValueError: If the string is not a valid image modality.
        """
        possible_str = super().from_str(value)
        if possible_str is not None:
            return possible_str

        raise ValueError(f"{value} is not a valid image modality")



# Copyright Justus Schock.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from operator import xor

import torch
from torchmetrics import Metric

__all__ = ["MovingAverage"]


class MovingAverage(Metric):
    sliding_window: list[torch.Tensor]
    current_average: torch.Tensor

    def __init__(
        self,
        momentum: float | None = 0.93,
        sliding_window_size: int | None = None,
    ) -> None:
        super().__init__()

        # need to add states here globally independent of arguments for momentum and sliding_window_size to satisfy mypy
        self.add_state("sliding_window", [], persistent=True)
        self.add_state("current_average", torch.tensor(float("nan")), persistent=True)

        assert xor(
            bool(sliding_window_size is None), bool(momentum is None)
        ), "Either momentum or sliding_window_size must be specified"
        if momentum is not None:
            if not 0.0 < momentum < 1.0:
                raise ValueError("momentum must be between 0.0 and 1.0")

        else:
            if sliding_window_size is None or not sliding_window_size > 0:
                raise ValueError("sliding_window_size must be greater than 0")

        self.momentum = momentum
        self.sliding_window_size = sliding_window_size

    def update(self, value: torch.Tensor) -> None:
        if self.momentum is not None:
            if torch.isnan(self.current_average):
                self.current_average = value.detach()

            else:
                self.current_average = (
                    self.momentum * self.current_average
                    + (1 - self.momentum) * value.detach()
                )
        else:
            self.sliding_window.append(value.detach())
            if self.sliding_window_size is None:
                raise ValueError("sliding_window_size must be specified")

            if len(self.sliding_window) > self.sliding_window_size:
                self.sliding_window.pop(0)

    def compute(self) -> torch.Tensor:
        if self.momentum is not None:
            return self.current_average
        else:
            result = sum(self.sliding_window) / len(self.sliding_window)
            if not isinstance(result, torch.Tensor):
                result = torch.tensor(result, device=self.device, dtype=torch.float)
            return result



# Copyright Justus Schock.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import gc
import math
from contextlib import contextmanager
from typing import Generator

import numpy as np
import torch
import torchio as tio
from pytorch_lightning import Trainer
from torch.utils.data import DataLoader

from pytorch_lightning import LightningModule
from mdlu.utils import suppress_stdout


def b2mb(x: float) -> float:
    return x / (2**20)


def mb2b(x: float) -> float:
    return x * (2**20)


class MemoryEstimator:
    def __init__(
        self, device_id: int = 0, mixed_precision: bool = False, offset: int = 1  # 1mb
    ) -> None:

        self.device_id = device_id
        self.device_props = torch.cuda.get_device_properties(device_id)
        self.offset = offset
        self.mixed_precision = mixed_precision

    @property
    def device_name(self) -> str:
        return self.device_props.name

    @property
    def total_memory(self) -> float:
        return b2mb(self.device_props.total_memory)

    def create_offset_tensor_on_gpu(self) -> torch.Tensor:
        device = torch.device(f"cuda:{self.device_id}")
        tensor_mem = torch.rand(
            1, dtype=float, requires_grad=False, device=device
        ).element_size()
        return torch.rand(
            math.ceil(self.offset / tensor_mem),
            dtype=float,
            requires_grad=False,
            device=device,
        )

    def __call__(
        self,
        network: LightningModule,
        batch_size: int,
        image_patch_shape: np.ndarray,
        output_shape: np.ndarray,
        whole_image_shape: np.ndarray,
        in_channels: int,
    ) -> tuple[int, bool]:

        try:
            device = torch.device(f"cuda:{self.device_id}")

            with cudnn_deterministic():
                torch.cuda.empty_cache()
                torch.cuda.reset_peak_memory_stats()
                network = network.to(device)

                block_tensor = self.create_offset_tensor_on_gpu().to(device=device)
                import time

                time.sleep(1)
                dset_train = [
                    {
                        "data": {
                            tio.constants.DATA: torch.rand(
                                (in_channels, *image_patch_shape),
                                dtype=torch.float,
                            )
                        },
                        "label": {
                            tio.constants.DATA: torch.zeros(
                                tuple(output_shape.tolist()),
                                dtype=torch.long,
                            ),
                        },
                    }
                    for i in range(batch_size * 10)
                ]

                dset_val = [
                    {
                        "data": {
                            tio.constants.DATA: torch.rand(
                                (in_channels, *whole_image_shape),
                                dtype=torch.float,
                            )
                        },
                        "label": {
                            tio.constants.DATA: torch.zeros(
                                tuple(output_shape.tolist()),
                                dtype=torch.long,
                            ),
                        },
                    }
                    for i in range(10)
                ]
                loader_train = DataLoader(
                    dset_train, batch_size=batch_size, shuffle=False
                )
                loader_val = DataLoader(dset_val, batch_size=1, shuffle=False)

                with suppress_stdout():
                    trainer = Trainer(
                        devices=[self.device_id],
                        accelerator="gpu",
                        precision=16 + int(not self.mixed_precision) * 16,
                        enable_model_summary=False,
                        enable_progress_bar=False,
                        max_epochs=1,
                        limit_train_batches=len(loader_train),
                        limit_val_batches=len(loader_val),
                        enable_checkpointing=False,
                        logger=None,
                    )
                    trainer.fit(network, loader_train, loader_val)

                torch.cuda.empty_cache()
                dyn_mem = torch.cuda.max_memory_reserved()

        except RuntimeError as e:
            if "CUDA out of memory" in str(e):
                dyn_mem = float("inf")
            else:
                raise e

        finally:
            for key in ("block_tensor", "trainer", "loader", "dset"):
                if key in locals():
                    del locals()[key]

            network.cpu()
            torch.cuda.empty_cache()
            gc.collect()

        return dyn_mem, torch.isfinite(torch.tensor(dyn_mem))


@contextmanager
def cudnn_deterministic() -> Generator[None, None, None]:
    old_value = torch.backends.cudnn.deterministic
    old_value_benchmark = torch.backends.cudnn.benchmark
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = True
    try:
        yield
    finally:
        torch.backends.cudnn.deterministic = old_value
        torch.backends.cudnn.benchmark = old_value_benchmark



# Copyright Justus Schock.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from copy import deepcopy
from typing import Any

import torch
from torchmetrics import Metric

__all__ = ["SingleClassMetric"]


class SingleClassMetric(Metric):
    def __init__(self, class_idx: int, metric_instance: Metric, **kwargs) -> None:
        super().__init__(**kwargs)
        self.metric = deepcopy(metric_instance)
        self.class_idx = class_idx

    def update(self, preds: torch.Tensor, target: torch.Tensor) -> None:
        preds = preds.argmax(1)
        preds = (preds == self.class_idx).long()
        target = (target == self.class_idx).long()

        return self.metric.update(preds, target)

    def compute(self) -> torch.Tensor | Any:
        return self.metric.compute()
